"""Helper functions for reading parameters of the model file."""

from array import array
from itertools import chain, repeat
import struct
from typing import BinaryIO, cast, Final, Iterator, Tuple

# Myanmar Unicode characters before digits
STD: Final = range(0x1000, 0x103F + 1)
# Myanmar Unicode characters after digits
AFT: Final = range(0x104A, 0x109F + 1)
# Myanmar Extended-A Unicode characters
EXA: Final = range(0xAA60, 0xAA7F + 1)
# Myanmar Extended-B Unicode characters
EXB: Final = range(0xA9E0, 0xA9FF + 1)
# Unicode space characters
SPC: Final = range(0x2000, 0x200B + 1)


def check_signature(stream: BinaryIO) -> str:
    """
    Check signature of the model file and return characters used by the model.

    The characters returned are sorted in lexicographical order.
    """
    uzmodel_tag = stream.read(8)
    if uzmodel_tag != b'UZMODEL ':
        raise IOError('invalid uzmodel_tag')
    uzmodel_version = read_int(stream)

    if uzmodel_version == 1:
        ssv = 0
    elif uzmodel_version == 2:
        ssv = read_int(stream)
    else:
        raise IOError('invalid uzmodel_version')

    if ssv == 0:
        chars = ''.join(map(chr, chain(STD, AFT, EXA, EXB, SPC)))
    elif ssv == 1:
        chars = ''.join(map(chr, chain(STD, AFT, EXA, EXB)))
    else:
        raise ValueError('invalid ssv')

    bmarkov_tag = stream.read(8)
    if bmarkov_tag != b'BMARKOV ':
        raise IOError('invalid bmarkov_tag')
    bmarkov_version = read_int(stream)
    if bmarkov_version != 0:
        raise IOError('invalid bmarkov_version')

    return chars


def read_params(stream: BinaryIO) -> 'array[float]':
    """Read parameters in the model file."""
    size = read_short(stream)
    params = array('f', repeat(0, size * size))
    for i in range(size):
        count = read_short(stream)
        if count != 0:
            offset = i * size
            # set default value
            value = read_float(stream)
            for index in range(size):
                params[offset + index] = value
            # set special values
            for index, value in read_pairs(stream, count):
                params[offset + index] = value
    return params


def read_short(stream: BinaryIO) -> int:
    """Read a short integer value in big-endian order."""
    return cast(int, struct.unpack('>h', stream.read(2))[0])


def read_int(stream: BinaryIO) -> int:
    """Read an integer value in big-endian order."""
    return cast(int, struct.unpack('>i', stream.read(4))[0])


def read_float(stream: BinaryIO) -> float:
    """Read a float value in big-endian order."""
    return cast(float, struct.unpack('>f', stream.read(4))[0])


def read_pairs(stream: BinaryIO, n: int) -> Iterator[Tuple[int, float]]:
    """Read n int-float value pairs in big-endian order."""
    return cast(
        Iterator[Tuple[int, float]],
        struct.iter_unpack('>hf', stream.read(6 * n))
    )



"""Zawgyi detector module."""

from bisect import bisect_left
from importlib.resources import open_binary
from itertools import chain, filterfalse
from math import exp, inf, isnan, nan
from typing import Iterator, Optional

from ._params import check_signature, read_params


class ZawgyiDetector:
    """A detector of Myanmar Zawgyi encoding."""

    __slots__ = ['_chars', '_params']

    def __init__(self) -> None:
        """Intialize the detector."""
        with open_binary(
            'myanmartools.resources',
            'zawgyiUnicodeModel.dat'
        ) as stream:
            self._chars = check_signature(stream)
            self._params = read_params(stream)
            # the 0 node is for foreign characters so mark as nan
            self._params[0] = nan

    def _state(self, char: Optional[str]) -> int:
        """
        Return the state of a character.

        Return 0 for foreign characters.
        """
        if char is None:
            return 0
        i = bisect_left(self._chars, char)
        if i < len(self._chars) and self._chars[i] == char:
            return i + 1
        return 0

    def _llrs(self, string: str) -> Iterator[float]:
        """
        Return the log-likelihood ratios of consecutive character pairs.

        The first and last characters are paired with None on the left
        and right respectively.
        """
        size = len(self._chars) + 1
        return map(
            lambda i, j: self._params[self._state(i) * size + self._state(j)],
            chain((None,), string),
            chain(string, (None,))
        )

    def get_zawgyi_probability(self, string: str) -> float:
        """
        Return the Zawgyi probability of a string.

        Return negative infinity if there are only foreign characters.
        """
        if all(map(isnan, self._llrs(string))):
            return -inf
        total = sum(filterfalse(isnan, self._llrs(string)))
        # Pz/(Pu+Pz) = exp(lnPz)/(exp(lnPu)+exp(lnPz)) = 1/(1+exp(lnPu-lnPz))
        # prevent overflow when positive
        if total >= 0:
            z = exp(-total)
            return z / (z + 1)
        return 1 / (1 + exp(total))



"""
Resources for myanmartools.

This peckage contains:

- `zawgyiUnicodeModel.dat` - parameters of Zawgyi detector
- `compatability.tsv` - Zawgyi probabilities and input strings
  resulting from Java implementation of Zawgyi detector
"""



"""Tests for Zawgyi detector module."""

import csv
from importlib.resources import open_text
import unittest

from myanmartools import ZawgyiDetector


class TestZawgyiDetector(unittest.TestCase):
    """Test for ZawgyiDetector class."""

    def setUp(self):
        """Create a ZawgyiDetector instance."""
        self.detector = ZawgyiDetector()

    def test_strong(self):
        """Test a Unicode string and a Zawgyi string."""
        strong_unicode = 'အပြည်ပြည်ဆိုင်ရာ လူ့အခွင့်အရေး ကြေညာစာတမ်း'
        strong_zawgyi = 'အျပည္ျပည္ဆိုင္ရာ လူ႔အခြင့္အေရး ေၾကညာစာတမ္း'
        self.assertLess(
            self.detector.get_zawgyi_probability(strong_unicode),
            0.001
        )
        self.assertGreater(
            self.detector.get_zawgyi_probability(strong_zawgyi),
            0.999
        )

    def test_compatibility(self):
        """Compare results with Java results."""
        with open_text('myanmartools.resources', 'compatibility.tsv') as f:
            for row in csv.reader(f, delimiter='\t'):
                self.assertAlmostEqual(
                    self.detector.get_zawgyi_probability(row[1]),
                    float(row[0])
                )
